
/*
 * Copyright (c) 2008 - 2009 NVIDIA Corporation.  All rights reserved.
 *
 * NVIDIA Corporation and its licensors retain all intellectual property and proprietary
 * rights in and to this software, related documentation and any modifications thereto.
 * Any use, reproduction, disclosure or distribution of this software and related
 * documentation without an express license agreement from NVIDIA Corporation is strictly
 * prohibited.
 *
 * TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED *AS IS*
 * AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS OR IMPLIED,
 * INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS BE LIABLE FOR ANY
 * SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES WHATSOEVER (INCLUDING, WITHOUT
 * LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS, BUSINESS INTERRUPTION, LOSS OF
 * BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS) ARISING OUT OF THE USE OF OR
 * INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGES
 */

#pragma once

#include <optixu/optixu_vector_types.h>
#include <optixu/optixu_math_namespace.h>
#include "./helpers.h"

typedef struct struct_BasicLight
{
#if defined(__cplusplus)
  typedef optix::float3 float3;
#endif
  float3 pos;
  float3 color;
  int    casts_shadow; 
  int    padding;      // make this structure 32 bytes -- powers of two are your friend!
} BasicLight;

struct TriangleLight
{
#if defined(__cplusplus)
  typedef optix::float3 float3;
#endif
  float3 v1, v2, v3;
  float3 normal;
  float3 emission;
  float uniformPdf;
};

struct TraditionalLight
{
#if defined(__cplusplus)
  typedef optix::float3 float3;
#endif
  float3 posOrDirection;
  float3 color;
  int type;     // 0 point light, 1 directional, 2 ambient (sum of all ambient terms from all lights, pos is invalid in that case)
  int padding;
};


struct MixedLight {
#if defined(__cplusplus)
  typedef optix::float3 float3;
#endif

    __device__ __host__ __inline__ void setAreaLight(const float3& v1, const float3& v2, const float3& v3, const float3& normal, const float3& emission) {
        this->v1 = v1;
        this->v2 = v2;
        this->v3 = v3;
        this->normal = optix::normalize(normal);
        this->emission = emission;
        this->area = optix::length(optix::cross(v1 - v2, v3 - v2)) * 0.5f;
    }

    __device__ __host__ __inline__ void setPointLight(const float3& pos, const float3& emission) {
        this->v1 = pos;
        this->v2 = optix::make_float3(0);
        this->v3 = optix::make_float3(0);
        this->normal = optix::make_float3(0);
        this->emission = emission;
        this->area = -1;
    }

    __device__ __host__ __inline__ void setDirectional(const float3& direction, const float3& emission) {
        this->v1 = optix::make_float3(0);
        this->v2 = optix::make_float3(0);
        this->v3 = optix::make_float3(0);
        this->normal = optix::normalize(direction);
        this->emission = emission;
        this->area = -2;
    }

    __device__ __host__ __inline__ float3 samplePosition(const float3& shadingPosition, float u, float v) { // u and v must be random numbers only if type == area (3)
        switch(type()) {
        case 0:
            return v1;
        case 1:
            return shadingPosition + normal * -100000.f;
        case 3:
        default:
            float w = 1.f - u - v;
            if(w<0.f) {
                w=1.f-u;
                u=1.f-v;
                v=w;
                w = 1.f - u - v;
            }
            return v1 * u + v2 * v + v3 * w;
        }
    }

    __device__ __host__ __inline__ char type() { // 0 point light, 1 directional, 3 area // removed ambient light, there will be a sky map for it and until now we have a separate miss colour
        if(area < -1.5)
            return 1;
        else if(area < -0.5)
            return 0;
        else return 3;
    }

    // pi in point lights, see http://blog.selfshadow.com/publications/s2013-shading-course/hoffman/s2013_pbs_physics_math_notes.pdf page 23
    // distance only point light and area light, cosThetaLight only area light
    __device__ __host__ __inline__ float3 incidentLight(float distance, float cosThetaLight) {
        switch(type()) {
        case 0:
            return emission * M_PIf / (distance * distance);
        case 1:
            return emission * M_PIf;
        case 3:
        default:
            return emission * cosThetaLight * area / (distance * distance);
        }
    }

    float3 normal;
private:
    float3 emission;
    float3 v1, v2, v3;
    float area;
};
