#include "Helsinki.h"

#include <optixu/optixpp_namespace.h>
#include <sutil.h>
#include <iostream>
#include <stdlib.h>
#include <string.h>
#include <GLUTDisplay.h>
#include <AssimpLoader.h>
#include <ObjLoader.h>
#include <sutil.h>
//#include <PPMLoader.h>
#include <sampleConfig.h>

#include "random.h"

using namespace optix;


namespace {
  std::string ptxpath( const std::string& base )
  {
    return std::string(sutilSamplesPtxDir()) + "/helsinki_generated_" + base + ".ptx";
  }
}

int myRTtimeoutcallback() {
    std::cout<<"timeout, try to reduce computation per frame, eg, decrease tile size in CleanScene::trace"<<std::endl;
    return 1;
}

Helsinki::Helsinki() :
    m_currentEntryPoint(DEFAULT_ENTRY_POINT),
    m_num_samples( 100u ),
    m_width(512u),
    m_height(512u)
{
    GLUTDisplay::setProgressiveSampleCount(m_num_samples);
}

void Helsinki::initScene( InitialCameraData& camera_data )
{
    m_context->setRayTypeCount( 1 );
    m_context->setEntryPointCount( 1 );
//    m_context->setStackSize( 1000 );

    m_context["scene_epsilon"]->setFloat( 1.e-3f );
    m_context["pathtrace_ray_type"]->setUint(0u);


    // Setup buffers
    Buffer buffer = createOutputBuffer( RT_FORMAT_FLOAT4, m_width, m_height );
    m_context["output_buffer"]->set(buffer);

    // Set up camera
    camera_data = InitialCameraData( make_float3( 0.0f, 0.0f, 0.5f ), // eye
                                   make_float3( 0.0f, 0.0f, 0.0f ),    // lookat
                                   make_float3( 0.0f, 1.0f,  0.0f ),       // up
                                   35.0f );                                // vfov

    // Declare these so validation will pass
    m_context["eye"]->setFloat( make_float3( 0.0f, 0.0f, 0.0f ) );
    m_context["U"]->setFloat( make_float3( 0.0f, 0.0f, 0.0f ) );
    m_context["V"]->setFloat( make_float3( 0.0f, 0.0f, 0.0f ) );
    m_context["W"]->setFloat( make_float3( 0.0f, 0.0f, 0.0f ) );

    m_context["bad_color"]->setFloat( 1.0f, 0.0f, 1.0f );
    m_context["bg_color"]->setFloat( make_float3(0.62f, 0.866f, 1.00f) );

    // Setup programs (path trace)
    std::string ptx_path = ptxpath( "helsinki", "BiDirCamera.cu" );
    Program exceptionProgram = m_context->createProgramFromPTXFile( ptx_path, "exception" );
    Program missProgram = m_context->createProgramFromPTXFile( ptx_path, "miss" );

    Program ray_gen_program = m_context->createProgramFromPTXFile( ptx_path, "pathtrace_camera" );
    m_context->setRayGenerationProgram(ENTRY_POINT_BIDIR_PATH_TRACE, ray_gen_program);
    m_context->setExceptionProgram(ENTRY_POINT_BIDIR_PATH_TRACE, exceptionProgram);


    m_context["frameNumber"]->setUint(1);
    m_context["sample_count"]->setInt(m_num_samples);

    // Create scene geometry

    createGeometry(camera_data, ptx_path);
    m_context->setMissProgram(0, missProgram);

    //rtContextSetTimeoutCallback()

    m_context->setTimeoutCallback(myRTtimeoutcallback, 1.8);
    // Finalize
    std::cout << "start validate" << std::endl;
    m_context->validate();
    std::cout << "start compile" << std::endl;
    m_context->compile();                                       // doesn't return
    std::cout << "end compile" << std::endl;
}

bool Helsinki::keyPressed( unsigned char key, int x, int y )
{
    return false;
}

void Helsinki::trace( const RayGenCameraData& camera_data )
{
    m_context["eye"]->setFloat( camera_data.eye );
    m_context["U"]->setFloat( camera_data.U );
    m_context["V"]->setFloat( camera_data.V );
    m_context["W"]->setFloat( camera_data.W );

    Buffer buffer = m_context["output_buffer"]->getBuffer();
    RTsize buffer_width, buffer_height;
    buffer->getSize( buffer_width, buffer_height );

    if( m_camera_changed ) {
        m_camera_changed = false;
        m_frame = 1;
    }

    m_context["frameNumber"]->setUint( m_frame );
    if(m_frame % 100 == 0)
        std::cout << "frameNumber: " << m_frame << std::endl;
    m_frame++;

    m_context->launch(m_currentEntryPoint, static_cast<unsigned int>(buffer_width), static_cast<unsigned int>(buffer_height));
}

//-----------------------------------------------------------------------------

Buffer Helsinki::getOutputBuffer()
{
  return m_context["output_buffer"]->getBuffer();
}

void Helsinki::createGeometry(InitialCameraData& camera_data, const std::string& ptx_path)
{
  // Set up material
  m_diffuseMaterial = m_context->createMaterial();

  m_closestHitProgram = m_context->createProgramFromPTXFile( ptx_path, "pathTracer" );


  m_diffuseMaterial->setClosestHitProgram( 0, m_closestHitProgram );


  // create geometry instances
  m_geometry_group = m_context->createGeometryGroup();

  m_geometry_group->setChildCount(0);
  m_geometry_group->setAcceleration(m_context->createAcceleration("Sbvh", "Bvh"));


  m_context["top_object"]->set( m_geometry_group );
}


