﻿#include "ValenciaPinholeCamera.h"
#include <optixu/optixu_matrix_namespace.h>

# ifdef _WIN32
#   include <cfloat>
#   define ISFINITE _finite
# else
#   include <cmath>
#   define ISFINITE std::isfinite
# endif

using namespace optix;

namespace {

/*
  Assigns src to dst if
  src is not inf and nan!

  dst = isReal(src) ? src : dst;
*/

float assignWithCheck( float& dst, const float &src )
{
  if( ISFINITE( src ) ) {
    dst = src;
  }

  return dst;
}

/*
  Assigns src to dst if all src
  components are neither inf nor nan!

  dst = isReal(src) ? src : dst;
*/

float3 assignWithCheck( float3& dst, const float3 &src )
{
  if( ISFINITE( src.x ) && ISFINITE( src.y ) && ISFINITE( src.z ) ) {
    dst = src;
  }

  return dst;
}

inline float DtoR(float d)
{
  return d*(static_cast<float>(M_PI)/180.f);
}


inline float RtoD(float r)
{
  return r*(180.f/static_cast<float>(M_PI));
}


inline float Clamp(float val, float min, float max)
{
  return val > min ? (val < max ? val : max) : min;
}

Matrix4x4 initWithBasis( const float3& u,
                               const float3& v,
                               const float3& w,
                               const float3& t )
{
  float m[16];
  m[0] = u.x;
  m[1] = v.x;
  m[2] = w.x;
  m[3] = t.x;

  m[4] = u.y;
  m[5] = v.y;
  m[6] = w.y;
  m[7] = t.y;

  m[8] = u.z;
  m[9] = v.z;
  m[10] = w.z;
  m[11] = t.z;

  m[12] = 0.0f;
  m[13] = 0.0f;
  m[14] = 0.0f;
  m[15] = 1.0f;

  return Matrix4x4( m );
}

inline float det3 (float a, float b, float c,
                   float d, float e, float f,
                   float g, float h, float i)
{ return a*e*i + d*h*c + g*b*f - g*e*c - d*b*i - a*h*f; }


/*
void printMatrix( const Matrix4x4& m )
{
  std::cerr << m[0] << " " << m[1] << " " << m[2] << " " << m[3] << "\n"
            << m[4] << " " << m[5] << " " << m[6] << " " << m[7] << "\n"
            << m[8] << " " << m[9] << " " << m[10] << " " << m[11] << "\n"
            << m[12] << " " << m[13] << " " << m[14] << " " << m[15] << std::endl;
}*/

#define mm(i,j) m[i*4+j]
float det4( const Matrix4x4& m )
{
  float det;
  det  = mm(0,0) * det3(mm(1,1), mm(1,2), mm(1,3),
                        mm(2,1), mm(2,2), mm(2,3),
                        mm(3,1), mm(3,2), mm(3,3));
  det -= mm(0,1) * det3(mm(1,0), mm(1,2), mm(1,3),
                        mm(2,0), mm(2,2), mm(2,3),
                        mm(3,0), mm(3,2), mm(3,3));
  det += mm(0,2) * det3(mm(1,0), mm(1,1), mm(1,3),
                        mm(2,0), mm(2,1), mm(2,3),
                        mm(3,0), mm(3,1), mm(3,3));
  det -= mm(0,3) * det3(mm(1,0), mm(1,1), mm(1,2),
                        mm(2,0), mm(2,1), mm(2,2),
                        mm(3,0), mm(3,1), mm(3,2));
  return det;
}

Matrix4x4 inverse( const Matrix4x4& m )
{
  Matrix4x4 inverse;
  float det = det4( m );

  inverse[0]  =  det3(mm(1,1), mm(1,2), mm(1,3),
                      mm(2,1), mm(2,2), mm(2,3),
                      mm(3,1), mm(3,2), mm(3,3)) / det;
  inverse[1]  = -det3(mm(0,1), mm(0,2), mm(0,3),
                      mm(2,1), mm(2,2), mm(2,3),
                      mm(3,1), mm(3,2), mm(3,3)) / det;
  inverse[2]  =  det3(mm(0,1), mm(0,2), mm(0,3),
                      mm(1,1), mm(1,2), mm(1,3),
                      mm(3,1), mm(3,2), mm(3,3)) / det;
  inverse[3]  = -det3(mm(0,1), mm(0,2), mm(0,3),
                      mm(1,1), mm(1,2), mm(1,3),
                      mm(2,1), mm(2,2), mm(2,3)) / det;

  inverse[4]  = -det3(mm(1,0), mm(1,2), mm(1,3),
                      mm(2,0), mm(2,2), mm(2,3),
                      mm(3,0), mm(3,2), mm(3,3)) / det;
  inverse[5]  =  det3(mm(0,0), mm(0,2), mm(0,3),
                      mm(2,0), mm(2,2), mm(2,3),
                      mm(3,0), mm(3,2), mm(3,3)) / det;
  inverse[6]  = -det3(mm(0,0), mm(0,2), mm(0,3),
                      mm(1,0), mm(1,2), mm(1,3),
                      mm(3,0), mm(3,2), mm(3,3)) / det;
  inverse[7]  =  det3(mm(0,0), mm(0,2), mm(0,3),
                      mm(1,0), mm(1,2), mm(1,3),
                      mm(2,0), mm(2,2), mm(2,3)) / det;

  inverse[8]  =  det3(mm(1,0), mm(1,1), mm(1,3),
                      mm(2,0), mm(2,1), mm(2,3),
                      mm(3,0), mm(3,1), mm(3,3)) / det;
  inverse[9]  = -det3(mm(0,0), mm(0,1), mm(0,3),
                      mm(2,0), mm(2,1), mm(2,3),
                      mm(3,0), mm(3,1), mm(3,3)) / det;
  inverse[10] =  det3(mm(0,0), mm(0,1), mm(0,3),
                      mm(1,0), mm(1,1), mm(1,3),
                      mm(3,0), mm(3,1), mm(3,3)) / det;
  inverse[11] = -det3(mm(0,0), mm(0,1), mm(0,3),
                      mm(1,0), mm(1,1), mm(1,3),
                      mm(2,0), mm(2,1), mm(2,3)) / det;

  inverse[12] = -det3(mm(1,0), mm(1,1), mm(1,2),
                      mm(2,0), mm(2,1), mm(2,2),
                      mm(3,0), mm(3,1), mm(3,2)) / det;
  inverse[13] =  det3(mm(0,0), mm(0,1), mm(0,2),
                      mm(2,0), mm(2,1), mm(2,2),
                      mm(3,0), mm(3,1), mm(3,2)) / det;
  inverse[14] = -det3(mm(0,0), mm(0,1), mm(0,2),
                      mm(1,0), mm(1,1), mm(1,2),
                      mm(3,0), mm(3,1), mm(3,2)) / det;
  inverse[15] =  det3(mm(0,0), mm(0,1), mm(0,2),
                      mm(1,0), mm(1,1), mm(1,2),
                      mm(2,0), mm(2,1), mm(2,2)) / det;

  return inverse;
}

#undef mm
}

//-----------------------------------------------------------------------------
//
// PinholeCamera definition
//
//-----------------------------------------------------------------------------


ValenciaPinholeCamera::ValenciaPinholeCamera(float3 eye, float3 lookat, float3 up, float hfov, float vfov, AspectRatioMode arm)
  : eye(eye)
  , lookat(lookat)
  , up(up)
  , hfov(hfov)
  , vfov(vfov)
  , aspectRatioMode(arm)
{
  setup();
}

void ValenciaPinholeCamera::setAspectRatio(float ratio)
{
  float realRatio = ratio;

  const float* inputAngle = 0;
  float* outputAngle = 0;
  switch(aspectRatioMode) {
  case ValenciaPinholeCamera::KeepHorizontal:
    inputAngle = &hfov;
    outputAngle = &vfov;
    realRatio = 1.f/ratio;
    break;
  case ValenciaPinholeCamera::KeepVertical:
    inputAngle = &vfov;
    outputAngle = &hfov;
    break;
  case ValenciaPinholeCamera::KeepNone:
    return;
    break;
  }

  *outputAngle = RtoD(2.0f*atan(realRatio*tan(DtoR(0.5f*(*inputAngle)))));

  setup();
}

void ValenciaPinholeCamera::setParameters(float3 eye_in, float3 lookat_in, float3 up_in, float hfov_in, float vfov_in, ValenciaPinholeCamera::AspectRatioMode aspectRatioMode_in)
{
  eye = eye_in;
  lookat = lookat_in;
  up = up_in;
  hfov = hfov_in;
  vfov = vfov_in;
  aspectRatioMode = aspectRatioMode_in;

  setup();
}

void ValenciaPinholeCamera::setup()
{
  lookdir = assignWithCheck( lookdir, lookat-eye );  // do not normalize lookdir -- implies focal length
  float lookdir_len = length( lookdir );
  up = assignWithCheck( up, normalize(up));
  camera_u = assignWithCheck( camera_u, normalize( cross(lookdir, up) ) );
  camera_v = assignWithCheck( camera_v, normalize( cross(camera_u, lookdir) ) );
  float ulen = lookdir_len * tan(DtoR(hfov*0.5f));
  camera_u = assignWithCheck( camera_u, camera_u * ulen );
  float vlen = lookdir_len * tan(DtoR(vfov*0.5f));
  camera_v = assignWithCheck( camera_v, camera_v * vlen );
}

void ValenciaPinholeCamera::getEyeUVW(float3& eye_out, float3& U, float3& V, float3& W)
{
  eye_out = eye;
  U = camera_u;
  V = camera_v;
  W = lookdir;
}

void ValenciaPinholeCamera::getEyeLookUpFOV(float3& eye_out, float3& lookat_out, float3& up_out, float& HFOV_out, float& VFOV_out)
{
  eye_out = eye;
  lookat_out = lookat;
  up_out = up;
  HFOV_out = hfov;
  VFOV_out = vfov;
}

void ValenciaPinholeCamera::scaleFOV(float scale)
{
  const float fov_min = 0.0f;
  const float fov_max = 120.0f;
  float hfov_new = RtoD(2.f*atan(scale*tan(DtoR(hfov*0.5f))));
  hfov_new = Clamp(hfov_new, fov_min, fov_max);
  float vfov_new = RtoD(2.f*atan(scale*tan(DtoR(vfov*0.5f))));
  vfov_new = Clamp(vfov_new, fov_min, fov_max);

  hfov = assignWithCheck( hfov, hfov_new );
  vfov = assignWithCheck( vfov, vfov_new );

  setup();
}

void ValenciaPinholeCamera::translate(float2 t)
{
	float3 trans = cross(up, normalize(lookdir)) * t.x + up * t.y;

	eye = assignWithCheck( eye, eye + trans );
	lookat = assignWithCheck( lookat, lookat + trans );

	setup();
}

void ValenciaPinholeCamera::translate(float t)
{
  float3 trans = normalize(lookdir) * t;

  eye = assignWithCheck( eye, eye + trans );
  lookat = assignWithCheck( lookat, lookat + trans );

  setup();
}

void ValenciaPinholeCamera::transform( const Matrix4x4& trans )
{
  float3 cen = eye;         // TODO: Add logic for various rotation types (eg, flythrough)

  Matrix4x4 frame = initWithBasis( normalize(camera_u),
                                         normalize(camera_v),
                                         normalize(-lookdir),
                                         cen );
  Matrix4x4 frame_inv = inverse( frame );

  Matrix4x4 final_trans = frame * trans * frame_inv;
  float4 up4     = make_float4( up, 0 );
  float4 eye4    = make_float4( eye, 0 );
  eye4.w         = 1.0f;
  float4 lookat4 = make_float4( lookat, 0 );
  lookat4.w      = 1.0f;


  up     = assignWithCheck( up, make_float3( final_trans*up4 ) );
  eye    = assignWithCheck( eye, make_float3( final_trans*eye4 ) );
  lookat = assignWithCheck( lookat, make_float3( final_trans*lookat4 ) );

  setup();
}
